/* 
** adduser 1.0: add a new user account (For systems not using shadow)
** With a nice little interface and a will to do all the work for you.
**
** Craig Hagan
** hagan@opine.cs.umass.edu
**
** Modified to really work, look clean, and find unused uid by Chris Cappuccio
** chris@slinky.cs.umass.edu
**
** MiNT port + adduser.conf support by Draco, 21.VI.1997.
**
** cc -O -o adduser adduser.c
** Use gcc if you have it... (political reasons beyond my control) (chris)
**
** I've gotten this program to work with success under Linux (without
** shadow) and SunOS 4.1.3. I would assume it should work pretty well
** on any system that uses no shadow. (chris)
**
** If you have no crypt() then try
** cc -DNO_CRYPT -O -o adduser adduser.c xfdes.c
** I'm not sure how login operates with no crypt()... I guess
** the same way we're doing it here.
*/

#include <pwd.h>
#include <grp.h>
#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/timeb.h>
#include <sys/time.h>
#include <sys/stat.h>

#define CONFIG_FILE	"/etc/adduser.conf"
#define PASSWD_FILE	"/etc/passwd"
#define DEFAULT_SHELL	"/bin/bash"
#define DEFAULT_HOME	"/home"
#define DEFAULT_MBOX	"/var/spool/mail"
#define DEFAULT_HTML	"public_html"
#define DEFAULT_GROUP	50	/* 50 is usually 'users' */
#define FIRST		400	/* First useable UID */
#define DEFAULT_PERMS	0755	/* Perms for the users home directory */

#ifndef MAXPATHLEN
#define MAXPATHLEN 128
#endif
#ifndef MAXLINE
#define MAXLINE 80
#endif

char passwd_path[MAXPATHLEN]=PASSWD_FILE;
char shell_path[32]=DEFAULT_SHELL;
char home_path[32]=DEFAULT_HOME;
char mbox_path[32]=DEFAULT_MBOX;
char html_dir[32]=DEFAULT_HTML;
int  init_group=DEFAULT_GROUP;
int  init_uid=FIRST;
int  init_perms=DEFAULT_PERMS;

int  mboxcreate=0,htmlcreate=0;

FILE *bull;	/* Bullshit way to find a shadow password file */
int unused_uid,nag;
char *crypt();

/* Draco: these functions are supposed to read the /etc/adduser.conf
 */

/* Octal string to integer conversion */

int otoi(char *string)
{
  int i, number = 0;

  for(i = 0; string[i] >= '0' && string[i] <= '7'; ++i)
	number = (number << 3) + string[i] - '0';
  return number;
}
  
setvar(char *name, char *line)
{
  char *key;

  if (key = strchr(line, '='))
      strcpy(name, ++key);
}
     
read_conf()
{
  FILE *fd;
  char line[MAXLINE];
  char *cp;
  char keypass[]="PASSWD";
  char keyshell[]="SHELL";
  char keyhome[]="HOME";
  char keygroup[]="GROUP";
  char keyfirst[]="FIRST_UID";
  char keyperms[]="PERMS";
  char keymbox[]="MAILBOX";
  char keyhtml[]="HTML";
  char keymbc[]="CREATE_MAILBOX";
  char keywww[]="CREATE_HTML";
  char *key;

  if ((fd = fopen(CONFIG_FILE, "r")) != NULL)
  {
     while(fgets(line, sizeof(line), fd) != NULL)
     {
        if (line[0] != '#')
	{
	  cp = strchr(line, '\n');
	  *cp = '\0';
	  if (strstr(line, keypass) != NULL)
          {
          	setvar(passwd_path, line);
                continue;
          }
	  if (strstr(line, keymbc) != NULL)
	  {
		key = strchr(line, '=');
		if (strstr(++key, "YES"))
			mboxcreate = 1;
		continue;
	  }
	  if (strstr(line, keywww) != NULL)
	  {
		key = strchr(line, '=');
		if (strstr(++key, "YES"))
			htmlcreate = 1;
		continue;
	  }
	  if (strstr(line, keyhtml) != NULL)
	  {
		setvar(html_dir, line);
		continue;
          }
	  if (strstr(line, keyshell) != NULL)
          {
		setvar(shell_path, line);
                continue;
          }
	  if (strstr(line, keyhome) != NULL)
	  {
		setvar(home_path, line);
                continue;
          }
	  if (strstr(line, keygroup) != NULL)
	  {
	    if (key = strchr(line, '=')) 
	    {
              	init_group = atoi(++key);
	      	continue;
	    }

	  }
	  if (strstr(line, keyfirst) != NULL)
	  {
            if (key = strchr(line, '='))
	    {	
              	init_uid = atoi(++key);
		continue;
	    }

          }
	  if (strstr(line, keyperms) != NULL)
          {
            if (key = strchr(line, '='))
	    {
              	init_perms = otoi(++key);
		continue;
	    }

          }
	  if (strstr(line, keymbox) != NULL)
	  {
	    if (key = strchr(line, '='))
		setvar(mbox_path, line);
 	  }

	}

     }
     (void) fclose(fd);
  }
/*
  printf("passwd: %s\nshell: %s\nhomedir: %s\ngroup: %d\nfirst: \
%d\nperms: %d", passwd_path, shell_path, home_path, init_group, \
init_uid, init_perms); 
 */
}

main()
{
  char foo[32];			

  /* a user read in from /etc/passwd this */
  /* to see if the username is in use     */				
  /* also for misc. inputs and other bullshit that i fell it should be */

  char uname[9],person[32],passwd[9],dir[32],shell[32],mbox[32],salt[2];
  char html_path[32],gethtml[32];

  /* uname=user name,person=full name,passwd=password,dir=home dir,
   * shell=shell,salt=for password encryption */

  unsigned int group,uid;

  /* the group and uid of the new user */

  int bad=0,done=0,correct=0,gets_warning=0,defaults=0,name_ok=0;

  /* flags, in order:
   * bad to see if the username is in /etc/passwd, or if strange stuff has
   * been typed if the user might be put in group 0
   * done allows the program to exit when a user has been added
   * correct loops until a password is found that isn't in /etc/passwd
   * gets_warning allows the fflush to be skipped for the first gets
   * so that output is still legible
   */

  time_t tm;

  struct passwd *pw;

  FILE *passwd_file;  /* Yep, it's a file allright */

  read_conf();
  
  /* set unused uid to FIRST (#defined above) so that find_unused picks
   * a uid over FIRST (assuming we do everything else right :)
   */

  unused_uid = init_uid;

  /* The real program starts HERE! */

  /* Smile, it's the 2nd best thing you can do with your lips */

  /* Lesse if we know what we're doing here... */
  
  if(geteuid()!=0)
  {
     printf("It seems you don't have access to add a new user.  Try\n");
     printf("logging in as root or su root to gain super-user access.\n");
     exit(1);
  }
  
  /* We don't support shadow password files, let's check and make
   * sure we're not using em. Sure we could use filesearch or something
   * but i don't feel like it.
   */

  if((bull=fopen("/etc/shadow","r"))!=NULL) /* SYSV Shadow */
  found_shadow();
  if((bull=fopen("/etc/master.passwd","r"))!=NULL) /* Hmm, BSD */
  found_shadow();

  /* The scent of SunOS NIS is in the air */

  if((bull=fopen("/var/yp","r"))!=NULL)
  {
  printf("You have Yellow Pages (NIS). Don't forget to 'make passwd' under /var/yp\n");
  printf("after you have made your new user.\n");
  fclose(bull);
  }

  while(!correct)		/* loop until a "good" uname is chosen */
    {				/* good = not in /etc/passwd */
      while(!done)
	{
	  name_ok=0;
	  while(!name_ok)
	  {
	  	printf("\nLogin to add (^C to quit): ");

	  	if(gets_warning)	/* if the warning was already shown */
	    		fflush(stdout);	/* fflush stdout, otherwise set the flag */
	  	else
	    		gets_warning=1;

	  	gets(uname);
		if (((uname[0]|0x20)>='a')&&((uname[0]|0x20)<='z'))
			name_ok=1;
		else
			printf("Invalid username, choose another.\n");
          }

          /* what I saw here before made me think maybe I was running DOS */
          /* might this be a solution? (chris) */
	  if(nag=getpwnam(uname) != NULL)
	    {
	      printf("That name is in use, choose another.\n");
	      done=0;
	    }
	  else done=1;
	}

      /* all set, get the rest of the stuff */

      printf("\nEditing information for new user [%s]\n",uname);
  
      printf("\nFull name for [%s]: ",uname);
      gets(person);

      defaults=0;
      printf("\nAutomatic mode? [Y/n]: ");
      gets(foo);
      defaults=(foo[0]=='N'||foo[0]=='n');

      if (defaults)
      {
      	printf("GID [%d]: ",init_group);
      	gets(foo);
      /* Draco: comparison instead of assignment... how did it work? */
        group=atoi(foo);
      	if (group==0)	/* You're not allowed to make root group users! */
		group=init_group;
      }
      else
	group=init_group;

      unused_uid = find_unused(++unused_uid);	/* our k-eleet unused! */

      if (defaults)
      { 
      	printf("\nUID [%d]: ",unused_uid);
      	gets(foo);
      	uid=atoi(foo);
      	if(uid==0) /* this is how i detect if you just hit return. */
        	uid=unused_uid;
                 /* it may also disable you from making a root user, but
                  * doesen't that sound more like a feature?
                  */
      }
      else
	uid=unused_uid;

      if((pw=getpwuid(uid))!=NULL)	/* uhh. duhh. egh.. */
      {
        printf("\nWarning: UID [%d] is already in use, this would conflict with\n",uid);
        printf("who is already owns that user ID. [%s]'s UID has been reset to\n",uname);
        printf("the last unused UID: [%d].\n",unused_uid);
        uid=unused_uid;
      }

      /* Draco: fix needed for possible next loop... */
      unused_uid--;
      
      fflush(stdin);

      if (defaults)
      {      
      	printf("\nHome Directory [%s/%s]: ",home_path,uname);
      	fflush(stdout);
      	gets(dir);
      	if (!strcmp(dir,""))
		sprintf(dir,"%s/%s",home_path,uname);
      	fflush(stdin);
      }
      else
	sprintf(dir,"%s/%s",home_path,uname);

      if ((uid >= init_uid) && htmlcreate)
      {
          if (defaults)
	  {
		printf("\nDefault HTML folder [%s]: ",html_dir); 
		fflush(stdout);
		gets(gethtml);
		if (!strcmp(gethtml,""))
			sprintf(html_path,"%s/%s/%s",home_path,uname,html_dir);
		else
			sprintf(html_path,"%s/%s/%s",home_path,uname,gethtml);
		fflush(stdin);
          }
          else
                sprintf(html_path,"%s/%s/%s",home_path,uname,html_dir);

       }
       if ((uid >= init_uid) && mboxcreate)
       {
	  if (defaults)
          { 
     		printf("\nMailbox directory [%s]: ",mbox_path);
      		fflush(stdout);
      		gets(foo);

      		if (!strcmp(foo,""))
			sprintf(mbox,"%s/%s",mbox_path,uname);
		else
			sprintf(mbox,"%s/%s",foo,uname);
      		fflush(stdin);
	  }
	  else
          	sprintf(mbox,"%s/%s",mbox_path,uname);
       }

      if (defaults)
      {	
      	printf("\nShell [%s]: ", shell_path);
      	fflush(stdout);
      	gets(shell);
      
      	if (!strcmp(shell,""))
		sprintf(shell,"%s", shell_path);
      }
      else
	sprintf(shell,"%s", shell_path);

      if (defaults)
      {
      	printf("\nPassword [%s]: ",uname);
      	fflush(stdout);
      	gets(passwd);

      	if (!strcmp(passwd,""))
		sprintf(passwd,"%s",uname);
      }
      else
	sprintf(passwd,"%s",uname);

      time(&tm);
      salt[0] = (tm & 0x0f) + 'A';
      salt[1] = ((tm & 0xf0) >> 4) + 'a';
  
      printf("\nInformation for new user [%s]:\n",uname);
      printf("Full name: %s\n", person);
      printf("Home directory: [%s] Shell: [%s]\n",dir,shell);
      if ((uid >= init_uid) && htmlcreate)
	printf("Default HTML folder: [%s]\n",html_path);
      if ((uid >= init_uid) && mboxcreate)
      	printf("Mailbox folder location: [%s]\n",mbox);
      printf("Password: [%s] uid: [%d] gid: [%d]\n",passwd,uid,group);
      printf("\nIs this correct? [y/N]: ");
      fflush(stdout);
      gets(foo);

      done=bad=correct=(foo[0]=='y'||foo[0]=='Y');

      if(bad!=1)
       printf("\nUser [%s] not added\n",uname);
    }

  printf("\nAdding user [%s]...\n",uname);
  mkdir(dir, init_perms);
  /* make sure that they own their directory -- its kinda nice :) */
  chown(dir,uid,group);

  if ((uid >= init_uid) && mboxcreate)
  {
  	creat(mbox,0600);
	chown(mbox,uid,group);
  }
  if ((uid >= init_uid) && htmlcreate)
  {
	mkdir(html_path,0755);
	chown(html_path,uid,group);
  }

  system("cp /etc/passwd /etc/opasswd"); /* Let's have safe sex */
  
  passwd_file=fopen(passwd_path,"a");

#ifdef NO_CRYPT
  fprintf(passwd_file,"%s:%s:%d:%d:%s:%s:%s\n"
  	  ,uname,fcrypt(passwd,salt),uid,group,person,dir,shell);
#else
  fprintf(passwd_file,"%s:%s:%d:%d:%s:%s:%s\n"
	  ,uname,crypt(passwd,salt),uid,group,person,dir,shell);
#endif

  fflush(passwd_file);
  fclose(passwd_file);
  /* yes, I fixed uid and group being screwed around (chris) */

}

/* Here is our trade secret patented copyrighted code to find an unused UID */

find_unused(begin)
	int begin;
{
	int trial;
	struct passwd *pw;
	trial = begin - 1;

        printf("\nChecking for an available UID after %d\n",init_uid);

	while ((pw = getpwuid(++trial)) != NULL)
	{
		printf("%d...",trial);
	}
	
	printf("\nFirst unused uid is %d\n", trial);

	return(trial);
}

found_shadow()
{
	printf("Sorry, it seems you are using some form of shadow password utilities\n");
	printf("instead of a normal password file which contains the actual encrypted\n");
	printf("password. You should use 'useradd' or equivalent user addition\n");
	printf("utility to make sure the new user is added correctly.\n");
	fclose(bull);
	exit(1);
}
